"""Flexible method to load confounds generated by fMRIprep.

Authors: load_confounds team
"""
import numpy as np
import pandas as pd
from . import confounds as cf
from .compcor import _find_compcor

# Global variables listing the admissible types of noise components
all_confounds = [
    "motion",
    "high_pass",
    "wm_csf",
    "global",
    "compcor",
    "ica_aroma",
    "scrub",
]


def _sanitize_strategy(strategy):
    """Defines the supported denoising strategies."""
    if isinstance(strategy, list):
        for conf in strategy:
            if not conf in all_confounds:
                raise ValueError(f"{conf} is not a supported type of confounds.")
    else:
        raise ValueError("strategy needs to be a list of strings")
    return strategy


def _check_error(missing_confounds, missing_keys):
    """Consolidate a single error message across multiple missing confounds."""
    if missing_confounds or missing_keys:
        error_msg = (
            "The following keys or parameters are missing: "
            + f" {missing_confounds}"
            + f" {missing_keys}"
            + ". You may want to try a different denoising strategy."
        )
        raise ValueError(error_msg)


class Confounds:
    """
    Confounds from fmriprep

    Parameters
    ----------
    strategy : list of strings
        The type of noise confounds to include.
        "motion" head motion estimates.
        "high_pass" discrete cosines covering low frequencies.
        "wm_csf" confounds derived from white matter and cerebrospinal fluid.
        "global" confounds derived from the global signal.
        "ica_aroma" confounds derived from ICA-AROMA.
        "scrub" regressors for Power 2014 scrubbing approach.

    motion : string, optional
        Type of confounds extracted from head motion estimates.
        "basic" translation/rotation (6 parameters)
        "power2" translation/rotation + quadratic terms (12 parameters)
        "derivatives" translation/rotation + derivatives (12 parameters)
        "full" translation/rotation + derivatives + quadratic terms + power2d derivatives (24 parameters)

    n_motion : float
        Number of pca components to keep from head motion estimates.
        If the parameters is strictly comprised between 0 and 1, a principal component
        analysis is applied to the motion parameters, and the number of extracted
        components is set to exceed `n_motion` percent of the parameters variance.
        If the n_components = 0, then no PCA is performed.

    fd_thresh : float, optional
        Framewise displacement threshold for scrub (default = 0.2 mm)

    std_dvars_thresh : float, optional
        Standardized DVARS threshold for scrub (default = 3)

    wm_csf : string, optional
        Type of confounds extracted from masks of white matter and cerebrospinal fluids.
        "basic" the averages in each mask (2 parameters)
        "power2" averages and quadratic terms (4 parameters)
        "derivatives" averages and derivatives (4 parameters)
        "full" averages + derivatives + quadratic terms + power2d derivatives (8 parameters)

    global_signal : string, optional
        Type of confounds extracted from the global signal.
        "basic" just the global signal (1 parameter)
        "power2" global signal and quadratic term (2 parameters)
        "derivatives" global signal and derivative (2 parameters)
        "full" global signal + derivatives + quadratic terms + power2d derivatives (4 parameters)

    scrub : string, optional
        Type of scrub of frames with excessive motion (Power et al. 2014)
        "basic" remove time frames based on excessive FD and DVARS
        "full" also remove time windows which are too short after scrubbing.
        one-hot encoding vectors are added as regressors for each scrubbed frame.

    compcor : string, optional
        Type of confounds extracted from a component based noise correction method
        "anat" noise components calculated using anatomical compcor
        "temp" noise components calculated using temporal compcor
        "full" noise components calculated using both temporal and anatomical

    n_compcor : int or "auto", optional
        The number of noise components to be extracted. For acompcor_combined=False,
        and/or compcor="full", this is the number of components per mask.
        Default is "auto": select all components (50% variance explained by fMRIPrep defaults)

    acompcor_combined: boolean, optional
        If true, use components generated from the combined white matter and csf
        masks. Otherwise, components are generated from each mask separately and then
        concatenated.

    ica_aroma : None or string, optional
        None: default, not using ICA-AROMA related strategy
        "basic": use noise IC only.
        "full": use fMRIprep output `~desc-smoothAROMAnonaggr_bold.nii.gz` .

    demean : boolean, optional
        If True, the confounds are standardized to a zero mean (over time).
        This step is critical if the confounds are regressed out of time series
        using nilearn with no or zscore standardization, but should be turned off
        with "spc" normalization.

    Attributes
    ----------
    `confounds_` : ndarray
        The confounds loaded using the specified model

    `columns_`: list of str
        The labels of the different confounds

    Notes
    -----
    The predefined strategies implemented in this class are
    adapted from (Ciric et al. 2017). Band-pass filter is replaced
    by high-pass filter. Low-pass filters can be implemented, e.g., through
    nilearn maskers. Scrubbing is implemented by introducing regressors in the
    confounds, rather than eliminating time points. Other aspects of the
    preprocessing listed in Ciric et al. (2017) are controlled through fMRIprep,
    e.g. distortion correction.

    References
    ----------
    Ciric et al., 2017 "Benchmarking of participant-level confound regression
    strategies for the control of motion artifact in studies of functional
    connectivity" Neuroimage 154: 174-87
    https://doi.org/10.1016/j.neuroimage.2017.03.020
    """

    def __init__(
        self,
        strategy=["motion", "high_pass", "wm_csf"],
        motion="full",
        n_motion=0,
        scrub="full",
        fd_thresh=0.2,
        std_dvars_thresh=3,
        wm_csf="basic",
        global_signal="basic",
        compcor="anat",
        acompcor_combined=True,
        n_compcor="auto",
        ica_aroma=None,
        demean=True,
    ):
        """Default parameters."""
        self.strategy = _sanitize_strategy(strategy)
        self.motion = motion
        self.n_motion = n_motion
        self.scrub = scrub
        self.fd_thresh = fd_thresh
        self.std_dvars_thresh = std_dvars_thresh
        self.wm_csf = wm_csf
        self.global_signal = global_signal
        self.compcor = compcor
        self.acompcor_combined = acompcor_combined
        self.n_compcor = n_compcor
        self.ica_aroma = ica_aroma
        self.demean = demean

    def load(self, img_files):
        """
        Load fMRIprep confounds

        Parameters
        ----------
        img_files : path to processed image files, optionally as a list.
            Processed nii.gz/dtseries.nii/func.gii file from fmriprep.
            `nii.gz` or `dtseries.nii`: path to files, optionally as a list.
            `func.gii`: list of a pair of paths to files, optionally as a list of lists.
            The companion tsv will be automatically detected.

        Returns
        -------
        confounds :  ndarray or list of ndarray
            A reduced version of fMRIprep confounds based on selected strategy and flags.
            An intercept is automatically added to the list of confounds.
        """
        img_files, flag_single = cf._sanitize_confounds(img_files)
        confounds_out = []
        columns_out = []
        self.missing_confounds_ = []
        self.missing_keys_ = []

        for file in img_files:
            conf, col = self._load_single(file)
            confounds_out.append(conf)
            columns_out.append(col)

        # If a single input was provided,
        # send back a single output instead of a list
        if flag_single:
            confounds_out = confounds_out[0]
            columns_out = columns_out[0]

        self.confounds_ = confounds_out
        self.columns_ = columns_out
        return confounds_out

    def _load_single(self, confounds_raw):
        """Load a single confounds file from fmriprep."""
        # Convert tsv file to pandas dataframe
        # check if relevant imaging files are present according to the strategy
        flag_acompcor = ("compcor" in self.strategy) and (self.compcor == "anat")
        flag_full_aroma = ("ica_aroma" in self.strategy) and (self.ica_aroma == "full")
        confounds_raw, self.json_ = cf._confounds_to_df(
            confounds_raw, flag_acompcor, flag_full_aroma
        )

        confounds = pd.DataFrame()

        for confound in self.strategy:
            loaded_confounds = self._load_confound(confounds_raw, confound)
            confounds = pd.concat([confounds, loaded_confounds], axis=1)

        _check_error(self.missing_confounds_, self.missing_keys_)
        confounds, labels = cf._confounds_to_ndarray(confounds, self.demean)
        return confounds, labels

    def _load_confound(self, confounds_raw, confound):
        """Load a single type of confound."""
        try:
            loaded_confounds = getattr(self, f"_load_{confound}")(confounds_raw)
        except cf.MissingConfound as exception:
            self.missing_confounds_ += exception.params
            self.missing_keys_ += exception.keywords
            loaded_confounds = pd.DataFrame()
        return loaded_confounds

    def _load_motion(self, confounds_raw):
        """Load the motion regressors."""
        motion_params = cf._add_suffix(
            ["trans_x", "trans_y", "trans_z", "rot_x", "rot_y", "rot_z"], self.motion
        )
        cf._check_params(confounds_raw, motion_params)
        confounds_motion = confounds_raw[motion_params]

        # Optionally apply PCA reduction
        if self.n_motion > 0:
            confounds_motion = cf._pca_motion(
                confounds_motion, n_components=self.n_motion
            )
        return confounds_motion

    def _load_high_pass(self, confounds_raw):
        """Load the high pass filter regressors."""
        high_pass_params = cf._find_confounds(confounds_raw, ["cosine"])
        return confounds_raw[high_pass_params]

    def _load_wm_csf(self, confounds_raw):
        """Load the regressors derived from the white matter and CSF masks."""
        wm_csf_params = cf._add_suffix(["csf", "white_matter"], self.wm_csf)
        cf._check_params(confounds_raw, wm_csf_params)
        return confounds_raw[wm_csf_params]

    def _load_global(self, confounds_raw):
        """Load the regressors derived from the global signal."""
        global_params = cf._add_suffix(["global_signal"], self.global_signal)
        cf._check_params(confounds_raw, global_params)
        return confounds_raw[global_params]

    def _load_compcor(self, confounds_raw):
        """Load compcor regressors."""
        compcor_cols = _find_compcor(
            self.json_, self.compcor, self.n_compcor, self.acompcor_combined
        )
        cf._check_params(confounds_raw, compcor_cols)
        return confounds_raw[compcor_cols]

    def _load_ica_aroma(self, confounds_raw):
        """Load the ICA-AROMA regressors."""
        if self.ica_aroma is None:
            raise ValueError("Please select an option when using ICA-AROMA strategy")
        if self.ica_aroma == "full":
            return pd.DataFrame()
        if self.ica_aroma == "basic":
            ica_aroma_params = cf._find_confounds(confounds_raw, ["aroma"])
            return confounds_raw[ica_aroma_params]

    def _load_scrub(self, confounds_raw):
        """Perform basic scrub - Remove volumes if framewise displacement exceeds threshold."""
        n_scans = len(confounds_raw)
        # Get indices of fd outliers
        fd_outliers = np.where(
            confounds_raw["framewise_displacement"] > self.fd_thresh
        )[0]
        dvars_outliers = np.where(confounds_raw["std_dvars"] > self.std_dvars_thresh)[0]
        combined_outliers = np.sort(
            np.unique(np.concatenate((fd_outliers, dvars_outliers)))
        )
        # Do full scrubbing if desired, and motion outliers were detected
        if self.scrub == "full" and len(combined_outliers) > 0:
            combined_outliers = cf._optimize_scrub(combined_outliers, n_scans)
        # Make one-hot encoded motion outlier regressors
        motion_outlier_regressors = pd.DataFrame(
            np.transpose(np.eye(n_scans)[combined_outliers]).astype(int)
        )
        column_names = [
            "motion_outlier_" + str(num)
            for num in range(np.shape(motion_outlier_regressors)[1])
        ]
        motion_outlier_regressors.columns = column_names
        return motion_outlier_regressors



"""loading fMRIprep confounds into python."""
from load_confounds.parser import Confounds
from load_confounds.strategies import (
    Minimal,
    Scrubbing,
    CompCor,
    ICAAROMA,
)

__all__ = [
    "Confounds",
    "Minimal",
    "Scrubbing",
    "CompCor",
    "ICAAROMA",
]



"""Helper function for _load_compcor."""


prefix_compcor = {"full": ["t", "a"], "temp": ["t"], "anat": ["a"]}
anat_masker = {True: ["combined"], False: ["WM", "CSF"], None: None}


def _find_compcor(confounds_json, compcor, n_compcor, acompcor_combined):
    """Builds list for the number of compcor components."""
    prefix_set, anat_mask = _check_compcor_method(compcor, acompcor_combined)

    collector = []
    for prefix in prefix_set:
        # all possible compcor confounds in order, mixing different types of mask
        all_compcor_name = [
            comp for comp in confounds_json.keys() if f"{prefix}_comp_cor" in comp
        ]
        # filter by prefix first (anat vs temp)
        compcor_cols_filt = _prefix_confound_filter(prefix, all_compcor_name)
        if prefix == "a":
            # apply acompor mask option if relevant, and select top components
            compcor_cols_filt = _acompcor_mask(
                confounds_json, anat_mask, compcor_cols_filt, n_compcor
            )
        else:
            # select top components
            compcor_cols_filt = _select_compcor(compcor_cols_filt, n_compcor)
        # Aggregate components across all masks
        collector += compcor_cols_filt
    return collector


def _select_compcor(compcor_cols, n_compcor):
    """Retain a specified number of compcor components."""
    # only select if not "auto", or less components are requested than there actually is
    if (n_compcor != "auto") and (n_compcor < len(compcor_cols)):
        compcor_cols = compcor_cols[0:n_compcor]
    return compcor_cols


def _check_compcor_method(compcor, acompcor_combined):
    """Load compcor options and check if method is acceptable."""
    # get relevant prefix from compcor strategy
    prefix_set = prefix_compcor[compcor]
    # get relevant compcore mask
    anat_mask = anat_masker[acompcor_combined]
    if ("a" in prefix_set) and (anat_mask is None):
        raise ValueError(
            f"acompcor_combined must set to True or False. Got {acompcor_combined}"
        )
    return prefix_set, anat_mask


def _acompcor_mask(confounds_json, anat_mask, compcor_cols_filt, n_compcor):
    """Filter according to acompcor mask(s) and select top components."""
    collector = []
    for mask in anat_mask:
        cols = _json_mask(compcor_cols_filt, confounds_json, mask)
        cols = _select_compcor(cols, n_compcor)
        collector += cols
    return collector


def _json_mask(compcor_cols_filt, confounds_json, mask):
    """Extract anat compcor components from a given mask."""
    cols = []
    for compcor_col in compcor_cols_filt:
        if confounds_json[compcor_col]["Mask"] in mask:
            cols.append(compcor_col)
    return cols


def _prefix_confound_filter(prefix, all_compcor_name):
    """Get confound columns by prefix and acompcor mask."""
    compcor_cols_filt = []
    for nn in range(len(all_compcor_name)):
        nn_str = str(nn).zfill(2)
        compcor_col = f"{prefix}_comp_cor_{nn_str}"
        compcor_cols_filt.append(compcor_col)
    return compcor_cols_filt



"""Test predefined denoising strategies."""
import os
import re
import load_confounds.strategies as lc
import numpy as np
import pytest


path_data = os.path.join(os.path.dirname(lc.__file__), "data")
file_confounds = os.path.join(
    path_data, "test_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
)
file_aroma = os.path.join(
    path_data, "test_space-MNI152NLin2009cAsym_desc-smoothAROMAnonaggr_bold.nii.gz"
)


def test_Minimal():
    """Test the Minimal strategy."""
    # Try to load the confounds, whithout PCA reduction
    conf = lc.Minimal()
    assert conf.strategy == ["high_pass", "motion", "wm_csf"]
    assert hasattr(conf, "global_signal") == False
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    # Check that all model categories have been successfully loaded
    list_check = [
        "trans_y",
        "trans_z",
        "rot_z",
        "cosine00",
        "csf",
        "white_matter",
    ]
    for check in list_check:
        assert check in conf.columns_

    # maker sure global signal works
    conf = lc.Minimal(global_signal="basic")
    assert conf.strategy == ["high_pass", "motion", "wm_csf", "global"]
    assert conf.global_signal == "basic"


def test_Scrubbing():
    """Test the Scrubbing strategy."""
    conf = lc.Scrubbing(fd_thresh=0.15)
    # make sure global signal is not there
    assert conf.strategy == ["high_pass", "motion", "wm_csf", "scrub"]
    assert hasattr(conf, "global_signal") == False
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    # Check that all model categories have been successfully loaded
    list_check = [
        "trans_x",
        "trans_y",
        "rot_z",
        "trans_x_derivative1",
        "trans_x_power2",
        "trans_x_derivative1_power2",
        "trans_y_derivative1",
        "trans_y_power2",
        "trans_y_derivative1_power2",
        "trans_z_derivative1",
        "trans_z_power2",
        "rot_z_derivative1",
        "rot_z_power2",
        "rot_z_derivative1_power2",
        "cosine00",
        "cosine01",
        "csf",
        "white_matter",
        "csf_derivative1",
        "csf_power2",
        "csf_derivative1_power2",
        "white_matter_derivative1",
        "motion_outlier_0",
        "motion_outlier_1",
    ]

    for check in list_check:
        assert check in conf.columns_

    # also load confounds with very liberal scrubbing thresholds
    # this should not produce an error
    conf = lc.Scrubbing(fd_thresh=1, std_dvars_thresh=5)
    conf.load(file_confounds)
    assert "motion_outlier_0" not in conf.columns_

    # maker sure global signal works
    conf = lc.Scrubbing(global_signal="full")
    assert conf.strategy == ["high_pass", "motion", "wm_csf", "scrub", "global"]
    assert conf.global_signal == "full"



def test_CompCor_anatomical():
    """Test the anatomical CompCor strategy."""
    # Try to load the confounds, whithout PCA reduction
    conf = lc.CompCor()
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    list_check = [
        "trans_x",
        "trans_y",
        "rot_z",
        "trans_x_derivative1",
        "trans_x_power2",
        "trans_y_derivative1_power2",
        "trans_z_derivative1",
        "trans_z_power2",
        "trans_z_derivative1_power2",
        "rot_y_derivative1",
        "rot_y_power2",
        "rot_z_power2",
        "rot_z_derivative1_power2",
        "cosine00",
        "cosine01",
        "a_comp_cor_00",
        "a_comp_cor_01",
        "a_comp_cor_02",
    ]

    for check in list_check:
        assert check in conf.columns_

    compcor_col_str_anat = "".join(conf.columns_)
    assert "t_comp_cor_" not in compcor_col_str_anat
    assert (
        "a_comp_cor_57" not in compcor_col_str_anat
    )  # this one comes from the WW mask


def test_CompCor_anatomical_not_combined():
    """Test the anatomical CompCor strategy without combined mask."""
    # Try to load the confounds, whithout PCA reduction
    conf = lc.CompCor(acompcor_combined=False, n_compcor=5)
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    list_check = [
        "trans_x",
        "trans_y",
        "rot_z",
        "trans_x_derivative1",
        "trans_x_power2",
        "trans_y_derivative1_power2",
        "trans_z_derivative1",
        "trans_z_power2",
        "trans_z_derivative1_power2",
        "rot_y_derivative1",
        "rot_y_power2",
        "rot_z_power2",
        "rot_z_derivative1_power2",
        "cosine00",
        "cosine01",
        "a_comp_cor_57",  # from CSF mask
        "a_comp_cor_61",  # from CSF mask
        "a_comp_cor_70",  # from WM mask
        "a_comp_cor_74",  # from WM mask
    ]

    for check in list_check:
        assert check in conf.columns_

    compcor_col_str_anat = "".join(conf.columns_)
    assert "t_comp_cor_" not in compcor_col_str_anat
    assert (
        "a_comp_cor_00" not in compcor_col_str_anat
    )  # this one comes from the combined mask
    assert (
        "a_comp_cor_62" not in compcor_col_str_anat
    )  # this one exceeds the number of requested components
    assert (
        "a_comp_cor_75" not in compcor_col_str_anat
    )  # this one exceeds the number of requested components


def test_CompCor_temporal():
    """Test the temporal ompCor strategy."""
    # Try to load the confounds, whithout PCA reduction
    conf = lc.CompCor(compcor="temp")
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    list_check = [
        "cosine00",
        "cosine01",
        "cosine02",
        "cosine03",
        "t_comp_cor_00",
        "t_comp_cor_01",
        "t_comp_cor_02",
        "t_comp_cor_03",
    ]
    for check in list_check:
        assert check in conf.columns_

    compcor_col_str_anat = "".join(conf.columns_)
    assert "a_comp_cor_" not in compcor_col_str_anat


def test_FullCompCor():
    """Test a full compcor strategy."""
    conf = lc.CompCor(compcor="full", acompcor_combined=False)
    conf.load(file_confounds)

    assert isinstance(conf.confounds_, np.ndarray)

    list_check = [
        "t_comp_cor_00",
        "t_comp_cor_01",
        "t_comp_cor_02",
        "t_comp_cor_03",
        "a_comp_cor_57",  # from CSF mask
        "a_comp_cor_58",  # from CSF mask
        "a_comp_cor_105",  # from WM mask
    ]
    for check in list_check:
        assert check in conf.columns_


def test_ICAAROMA():
    """Test the (non-aggressive) ICA-AROMA strategy."""
    conf = lc.ICAAROMA(global_signal="basic")
    assert conf.global_signal == "basic"
    assert conf.strategy == ["wm_csf", "high_pass", "ica_aroma", "global"]
    conf.load(file_aroma)

    # Check that all fixed name model categories have been successfully loaded
    list_check = [
        "csf",
        "white_matter",
        "global_signal",
    ]
    for c in conf.columns_:
        # Check that all fixed name model categories
        fixed = c in list_check
        cosines = re.match("cosine+", c)
        assert fixed or (cosines is not None)


def test_invalid():
    """Test warning raised for invalid keywors."""
    with pytest.warns(UserWarning) as record:
        lc.ICAAROMA(compcor="anat", global_signal="full")
    assert "not taking effect: ['compcor']" in record[0].message.args[0]



"""Predefined denoising strategies.

Authors: load_confounds team
"""
import warnings
from .parser import Confounds


class Minimal(Confounds):
    """
    Load confounds for a minimal denosing strategy commonly used
    in resting state functional connectivity, described in Fox et al., 2005.
    Full motion parameters, WM/CSF signals, and high pass filter,
    with an option to extract global signal confounds.

    Parameters
    ----------
    confounds_raw : Pandas Dataframe or path to tsv file(s), optionally as a list.
        Raw confounds from fmriprep

    motion : string, optional
        Type of confounds extracted from head motion estimates.
        "basic" translation/rotation (6 parameters)
        "power2" translation/rotation + quadratic terms (12 parameters)
        "derivatives" translation/rotation + derivatives (12 parameters)
        "full" translation/rotation + derivatives + quadratic terms + power2d derivatives (24 parameters)

    wm_csf : string, optional
        Type of confounds extracted from masks of white matter and cerebrospinal fluids.
        "basic" the averages in each mask (2 parameters)
        "power2" averages and quadratic terms (4 parameters)
        "derivatives" averages and derivatives (4 parameters)
        "full" averages + derivatives + quadratic terms + power2d derivatives (8 parameters)

    demean : boolean, optional
        If True, the confounds are standardized to a zero mean (over time).
        This step is critical if the confounds are regressed out of time series
        using nilearn with no or zscore standardization, but should be turned off
        with "spc" normalization.

    global_signal : string, optional
        Specify type of confounds extracted from the global signal.
        Global signal regressors will not be retrieved if no arguments were applied.
        "basic" just the global signal (1 parameter)
        "power2" global signal and quadratic term (2 parameters)
        "derivatives" global signal and derivative (2 parameters)
        "full" global signal + derivatives + quadratic terms + power2d derivatives (4 parameters)

    Returns
    -------
    conf :  a Confounds object
        conf.confounds_ is a reduced version of fMRIprep confounds.

    """

    def __init__(self, motion="full", wm_csf="basic", demean=True, **kwargs):
        """Default parameters."""
        # check if global signal is supplied as a parameter
        # if so, add to strategy
        global_signal = kwargs.get("global_signal", False)
        strategy = ["high_pass", "motion", "wm_csf"]
        strategy, global_signal = _update_strategy(strategy, global_signal)
        # warn user for supplying useless parameter
        _check_invalid_parameter(kwargs, valid_keys=["global_signal"])

        # set attributes
        self.strategy = strategy
        self.motion = motion
        self.n_motion = 0
        self.wm_csf = wm_csf
        self.demean = demean
        if global_signal:
            self.global_signal = global_signal


class Scrubbing(Confounds):
    """
    Load confounds for scrubbing describbed in Power et al., 2012.
    Motion parameters, WM/CSF signals, scrub (full), high pass filter.
    All noise components are fully expanded (derivatives, squares and squared
    derivatives).

    Parameters
    ----------
    confounds_raw : Pandas Dataframe or path to tsv file(s), optionally as a list.
        Raw confounds from fmriprep

    motion : string, optional
        Type of confounds extracted from head motion estimates.
        "basic" translation/rotation (6 parameters)
        "power2" translation/rotation + quadratic terms (12 parameters)
        "derivatives" translation/rotation + derivatives (12 parameters)
        "full" translation/rotation + derivatives + quadratic terms + power2d derivatives (24 parameters)

    wm_csf : string, optional
        Type of confounds extracted from masks of white matter and cerebrospinal fluids.
        "basic" the averages in each mask (2 parameters)
        "power2" averages and quadratic terms (4 parameters)
        "derivatives" averages and derivatives (4 parameters)
        "full" averages + derivatives + quadratic terms + power2d derivatives (8 parameters)

    scrub : string, optional
        Type of scrub of frames with excessive motion (Power et al. 2014)
        "basic" remove time frames based on excessive FD and DVARS
        "full" also remove time windows which are too short after scrubbing.
        one-hot encoding vectors are added as regressors for each scrubbed frame.

    fd_thresh : float, optional
        Framewise displacement threshold for scrub (default = 0.2 mm)

    std_dvars_thresh : float, optional
        Standardized DVARS threshold for scrub (default = 3)

    demean : boolean, optional
        If True, the confounds are standardized to a zero mean (over time).
        This step is critical if the confounds are regressed out of time series
        using nilearn with no or zscore standardization, but should be turned off
        with "spc" normalization.

    global_signal : string, optional
        Specify type of confounds extracted from the global signal.
        Global signal regressors will not be retrieved if no arguments were applied.
        "basic" just the global signal (1 parameter)
        "power2" global signal and quadratic term (2 parameters)
        "derivatives" global signal and derivative (2 parameters)
        "full" global signal + derivatives + quadratic terms + power2d derivatives (4 parameters)

    Returns
    -------
    conf :  a Confounds object
        conf.confounds_ is a reduced version of fMRIprep confounds.

    """

    def __init__(
        self,
        motion="full",
        wm_csf="full",
        scrub="full",
        fd_thresh=0.2,
        std_dvars_thresh=3,
        demean=True,
        **kwargs,
    ):
        """Default parameters."""
        # check if global signal is supplied as a parameter
        # if so, add to strategy
        global_signal = kwargs.get("global_signal", False)
        strategy = ["high_pass", "motion", "wm_csf", "scrub"]
        strategy, global_signal = _update_strategy(strategy, global_signal)
        # warn user for supplying useless parameter
        _check_invalid_parameter(kwargs, valid_keys=["global_signal"])

        # set attributes
        self.strategy = strategy
        self.motion = motion
        self.n_motion = 0
        self.wm_csf = wm_csf
        self.scrub = scrub
        self.fd_thresh = (fd_thresh,)
        self.std_dvars_thresh = (std_dvars_thresh,)
        self.demean = demean
        if global_signal:
            self.global_signal = global_signal


class CompCor(Confounds):
    """
    Load confounds using the CompCor strategy from Behzadi et al., 2007.
    Default with motion parameters (fully expanded), high pass filter, and anatomical compcor.

    Parameters
    ----------
    confounds_raw : Pandas Dataframe or path to tsv file(s), optionally as a list.
        Raw confounds from fmriprep

    motion : string, optional
        Type of confounds extracted from head motion estimates.
        "basic" translation/rotation (6 parameters)
        "power2" translation/rotation + quadratic terms (12 parameters)
        "derivatives" translation/rotation + derivatives (12 parameters)
        "full" translation/rotation + derivatives + quadratic terms + power2d derivatives (24 parameters)

    n_compcor : int or "auto", optional
        The number of noise components to be extracted. For acompcor_combined=False,
        this is the number of components per mask.
        Default is "auto": select all components (50% variance explained by fMRIPrep defaults)

    acompcor_combined: boolean, optional
        If true, use components generated from the combined white matter and csf
        masks. Otherwise, components are generated from each mask separately and then
        concatenated.

    demean : boolean, optional
        If True, the confounds are standardized to a zero mean (over time).
        This step is critical if the confounds are regressed out of time series
        using nilearn with no or zscore standardization, but should be turned off
        with "spc" normalization.

    Returns
    -------
    conf :  a Confounds object
        conf.confounds_ is a reduced version of fMRIprep confounds.

    """

    def __init__(
        self,
        motion="full",
        compcor="anat",
        n_compcor="auto",
        demean=True,
        acompcor_combined=True,
    ):
        """Default parameters."""
        # set attributes
        self.strategy = ["high_pass", "motion", "compcor"]
        self.motion = motion
        self.n_motion = 0
        self.compcor = compcor
        self.n_compcor = n_compcor
        self.acompcor_combined = acompcor_combined
        self.demean = demean


class ICAAROMA(Confounds):
    """
    Load confounds for non-aggresive ICA-AROMA strategy from Pruim et al., 2015.
    The strategy requires fMRIprep outputs generated with `--use-aroma`.

    ICA-AROMA is implemented in two steps:
    1. A non-aggressive denoising immediately after ICA classification.
        A linear regression estimates signals with all independent components
        as predictors.
        A partial regression is then applied to remove variance associated
        with noise independent components.
        fMRIprep perfoms this step and generates files suffixed with
        `desc-smoothAROMAnonaggr_bold`.
    2. Confound regression step (mean signals from WM and CSF)
        Confound regressors generated by `load_confounds.ICAAROMA`.
        The generated confound regressors must only be used on fMRIprep output
        suffixed `desc-smoothAROMAnonaggr_bold`.

    `desc-smoothAROMAnonaggr_bold` is generated in `MNI152NLin6Asym` only.
    To produce `desc-smoothAROMAnonaggr_bold` in other spatial templates,
    use FSL function `fsl_regfilt`. For example, native T1w space:
    ```
    fsl_regfilt -i sub-<subject_label>_task-<task_id>_space-T1w_desc-preproc_bold.nii.gz \
        -f $(cat sub-<subject_label>_task-<task_id>_AROMAnoiseICs.csv) \
        -d sub-<subject_label>_task-<task_id>_desc-MELODIC_mixing.tsv \
        -o sub-<subject_label>_task-<task_id>_space-T1w_desc-AROMAnonaggr_bold.nii.gz
    ```

    Parameters
    ----------
    confounds_raw : Pandas Dataframe or path to tsv file(s), optionally as a list.
        Raw confounds from fmriprep

    wm_csf : string, optional
        Type of confounds extracted from masks of white matter and cerebrospinal fluids.
        "basic" the averages in each mask (2 parameters)
        "power2" averages and quadratic terms (4 parameters)
        "derivatives" averages and derivatives (4 parameters)
        "full" averages + derivatives + quadratic terms + power2d derivatives (8 parameters)

    demean : boolean, optional
        If True, the confounds are standardized to a zero mean (over time).
        This step is critical if the confounds are regressed out of time series
        using nilearn with no or zscore standardization, but should be turned off
        with "spc" normalization.

    global_signal : string, optional
        Specify type of confounds extracted from the global signal.
        Global signal regressors will not be retrieved if no arguments were applied.
        "basic" just the global signal (1 parameter)
        "power2" global signal and quadratic term (2 parameters)
        "derivatives" global signal and derivative (2 parameters)
        "full" global signal + derivatives + quadratic terms + power2d derivatives (4 parameters)


    Returns
    -------
    conf :  a Confounds object
        conf.confounds_ is a reduced version of fMRIprep confounds.

    Notes
    -----
    fMRIprep documentation on ICA-AROMA
    https://fmriprep.org/en/latest/workflows.html#ica-aroma

    For more discussion regarding choosing the nuisance regressors before or
    after denoising with ICA-AROMA has a detriment on outcome measures,
    please see notebook 5.
    https://github.com/nipreps/fmriprep-notebooks/

    """

    def __init__(self, wm_csf="basic", demean=True, **kwargs):
        """Default parameters."""
        strategy = ["wm_csf", "high_pass", "ica_aroma"]
        global_signal = kwargs.get("global_signal", False)
        strategy, global_signal = _update_strategy(strategy, global_signal)
        # warn user for supplying useless parameter
        _check_invalid_parameter(kwargs, valid_keys=["global_signal"])

        # set attributes
        self.strategy = strategy
        self.demean = demean
        self.wm_csf = wm_csf
        self.ica_aroma = "full"
        if global_signal:
            self.global_signal = global_signal


def _check_invalid_parameter(keyword_args, valid_keys):
    """Raise warnings if kwargs contains invalid parameters."""
    # supply extra parameter will not effect the behaviour
    # but it is good to inform the user
    for key in valid_keys:
        if isinstance(keyword_args, dict) and key in keyword_args:
            keyword_args.pop(key)
    if isinstance(keyword_args, dict) and len(keyword_args) > 0:
        warnings.warn(
            "Supplied paramerters not accepted in the current "
            "strategy, hence not taking effect: "
            f"{list(keyword_args.keys())}. "
            "Please consider customising strategy with using "
            "the `Confounds` module."
        )


def _update_strategy(strategy, global_signal):
    """Update strategy if global signal is supplied as a parameter."""
    strat = strategy.copy()
    if isinstance(global_signal, str):
        strat.append("global")
    return strat, global_signal



import os
import re
import load_confounds.parser as lc
import pandas as pd
import numpy as np
from scipy.stats import pearsonr
from sklearn.preprocessing import scale
import pytest
from nibabel import Nifti1Image
from nilearn.input_data import NiftiMasker


path_data = os.path.join(os.path.dirname(lc.__file__), "data")
file_confounds = os.path.join(
    path_data, "test_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
)


def _simu_img(demean=True):
    """Simulate an nifti image based on confound file with some parts confounds and some parts noise."""
    # set the size of the image matrix
    nx = 5
    ny = 5
    # the actual number of slices will actually be double of that
    # as we will stack slices with confounds on top of slices with noise
    nz = 2
    # Load a simple 6 parameters motion models as confounds
    X = lc.Confounds(strategy=["motion"], motion="basic", demean=demean).load(
        file_confounds
    )
    # repeat X in length (axis = 0) three times to increase the degree of freedom
    X = np.tile(X, (3, 1))

    # the number of time points is based on the example confound file
    nt = X.shape[0]
    # initialize an empty 4D volume
    vol = np.zeros([nx, ny, 2 * nz, nt])
    vol_conf = np.zeros([nx, ny, 2 * nz])
    vol_rand = np.zeros([nx, ny, 2 * nz])

    # create a random mixture of confounds
    # standardized to zero mean and unit variance
    beta = np.random.rand(nx * ny * nz, X.shape[1])
    tseries_conf = scale(np.matmul(beta, X.transpose()), axis=1)
    # fill the first half of the 4D data with the mixture
    vol[:, :, 0:nz, :] = tseries_conf.reshape(nx, ny, nz, nt)
    vol_conf[:, :, 0:nz] = 1

    # create random noise in the second half of the 4D data
    tseries_rand = scale(np.random.randn(nx * ny * nz, nt), axis=1)
    vol[:, :, range(nz, 2 * nz), :] = tseries_rand.reshape(nx, ny, nz, nt)
    vol_rand[:, :, range(nz, 2 * nz)] = 1

    # Shift the mean to non-zero
    vol = vol + 100

    # create an nifti image with the data, and corresponding mask
    img = Nifti1Image(vol, np.eye(4))
    mask_conf = Nifti1Image(vol_conf, np.eye(4))
    mask_rand = Nifti1Image(vol_rand, np.eye(4))

    return img, mask_conf, mask_rand, X


def _tseries_std(img, mask_img, confounds, standardize):
    """Get the std of time series in a mask."""
    masker = NiftiMasker(mask_img=mask_img, standardize=standardize)
    tseries = masker.fit_transform(img, confounds=confounds)
    return tseries.std(axis=0)


def _denoise(img, mask_img, confounds, standardize):
    """Extract time series with and without confounds."""
    masker = NiftiMasker(mask_img=mask_img, standardize=standardize)
    tseries_raw = masker.fit_transform(img)
    tseries_clean = masker.fit_transform(img, confounds=confounds)
    return tseries_raw, tseries_clean


def _corr_tseries(tseries1, tseries2):
    """Compute the correlation between two sets of time series."""
    corr = np.zeros(tseries1.shape[1])
    for ind in range(tseries1.shape[1]):
        corr[ind], _ = pearsonr(tseries1[:, ind], tseries2[:, ind])
    return corr


def _regression(confounds):
    """Simple regression with nilearn."""
    # Simulate data
    img, mask_conf, _, _ = _simu_img(demean=True)
    confounds = np.tile(confounds, (3, 1))  # matching L29 (_simu_img)

    # Do the regression
    masker = NiftiMasker(mask_img=mask_conf, standardize=True)
    tseries_clean = masker.fit_transform(img, confounds)
    assert tseries_clean.shape[0] == confounds.shape[0]


@pytest.mark.filterwarnings("ignore")
def test_nilearn_regress():
    """Try regressing out all motion types in nilearn."""
    # Regress full motion
    confounds = lc.Confounds(strategy=["motion"], motion="full").load(file_confounds)
    _regression(confounds)

    # Regress high_pass
    confounds = lc.Confounds(strategy=["high_pass"]).load(file_confounds)
    _regression(confounds)

    # Regress wm_csf
    confounds = lc.Confounds(strategy=["wm_csf"], wm_csf="full").load(file_confounds)
    _regression(confounds)
    # Regress global
    confounds = lc.Confounds(strategy=["global"], global_signal="full").load(
        file_confounds
    )
    _regression(confounds)

    # Regress AnatCompCor
    confounds = lc.Confounds(strategy=["compcor"], compcor="anat").load(file_confounds)
    _regression(confounds)

    # Regress TempCompCor
    confounds = lc.Confounds(strategy=["compcor"], compcor="temp").load(file_confounds)
    _regression(confounds)

    # Regress ICA-AROMA
    confounds = lc.Confounds(strategy=["ica_aroma"], ica_aroma="basic").load(
        file_confounds
    )
    _regression(confounds)


@pytest.mark.filterwarnings("ignore")
def test_nilearn_standardize_false():
    """Test removing confounds in nilearn with no standardization."""
    # Simulate data
    img, mask_conf, mask_rand, X = _simu_img(demean=True)

    # Check that most variance is removed
    # in voxels composed of pure confounds
    tseries_std = _tseries_std(img, mask_conf, X, False)
    assert np.mean(tseries_std < 0.0001)

    # Check that most variance is preserved
    # in voxels composed of random noise
    tseries_std = _tseries_std(img, mask_rand, X, False)
    assert np.mean(tseries_std > 0.9)


@pytest.mark.filterwarnings("ignore")
def test_nilearn_standardize_zscore():
    """Test removing confounds in nilearn with zscore standardization."""
    # Simulate data

    img, mask_conf, mask_rand, X = _simu_img(demean=True)

    # We now load the time series with vs without confounds
    # in voxels composed of pure confounds
    # the correlation before and after denoising should be very low
    # as most of the variance is removed by denoising
    tseries_raw, tseries_clean = _denoise(img, mask_conf, X, "zscore")
    corr = _corr_tseries(tseries_raw, tseries_clean)
    assert corr.mean() < 0.2

    # We now load the time series with zscore standardization
    # with vs without confounds in voxels where the signal is uncorrelated
    # with confounds. The correlation before and after denoising should be very
    # high as very little of the variance is removed by denoising
    tseries_raw, tseries_clean = _denoise(img, mask_rand, X, "zscore")
    corr = _corr_tseries(tseries_raw, tseries_clean)
    assert corr.mean() > 0.8


def test_nilearn_standardize_psc():
    """Test removing confounds in nilearn with psc standardization."""
    # Similar test to test_nilearn_standardize_zscore, but with psc
    # Simulate data

    img, mask_conf, mask_rand, X = _simu_img(demean=False)

    # Areas with confound
    tseries_raw, tseries_clean = _denoise(img, mask_conf, X, "psc")
    corr = _corr_tseries(tseries_raw, tseries_clean)
    assert corr.mean() < 0.2

    # Areas with random noise
    tseries_raw, tseries_clean = _denoise(img, mask_rand, X, "psc")
    corr = _corr_tseries(tseries_raw, tseries_clean)
    assert corr.mean() > 0.8


def test_confounds2df():
    """Check auto-detect of confonds from an fMRI nii image."""
    conf = lc.Confounds()
    file_confounds_nii = os.path.join(
        path_data, "test_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
    )
    conf.load(file_confounds_nii)
    assert "trans_x" in conf.columns_


def test_sanitize_strategy():
    """Check that flawed strategy options generate meaningful error messages."""
    with pytest.raises(ValueError):
        lc.Confounds(strategy="string")

    with pytest.raises(ValueError):
        lc.Confounds(strategy=["error"])

    with pytest.raises(ValueError):
        lc.Confounds(strategy=[0])


def test_motion():

    conf_basic = lc.Confounds(strategy=["motion"], motion="basic")
    conf_basic.load(file_confounds)
    conf_derivatives = lc.Confounds(strategy=["motion"], motion="derivatives")
    conf_derivatives.load(file_confounds)
    conf_power2 = lc.Confounds(strategy=["motion"], motion="power2")
    conf_power2.load(file_confounds)
    conf_full = lc.Confounds(strategy=["motion"], motion="full")
    conf_full.load(file_confounds)

    params = ["trans_x", "trans_y", "trans_z", "rot_x", "rot_y", "rot_z"]
    for param in params:
        # Basic 6 params motion model
        assert f"{param}" in conf_basic.columns_
        assert f"{param}_derivative1" not in conf_basic.columns_
        assert f"{param}_power2" not in conf_basic.columns_
        assert f"{param}_derivative1_power2" not in conf_basic.columns_

        # Use a 6 params + derivatives motion model
        assert f"{param}" in conf_derivatives.columns_
        assert f"{param}_derivative1" in conf_derivatives.columns_
        assert f"{param}_power2" not in conf_derivatives.columns_
        assert f"{param}_derivative1_power2" not in conf_derivatives.columns_

        # Use a 6 params + power2 motion model
        assert f"{param}" in conf_power2.columns_
        assert f"{param}_derivative1" not in conf_power2.columns_
        assert f"{param}_power2" in conf_power2.columns_
        assert f"{param}_derivative1_power2" not in conf_power2.columns_

        # Use a 6 params + derivatives + power2 + power2d derivatives motion model
        assert f"{param}" in conf_full.columns_
        assert f"{param}_derivative1" in conf_full.columns_
        assert f"{param}_power2" in conf_full.columns_
        assert f"{param}_derivative1_power2" in conf_full.columns_


def test_n_compcor():

    conf = lc.Confounds(strategy=["compcor"], compcor="anat", n_compcor=2)
    conf.load(file_confounds)
    assert "a_comp_cor_00" in conf.columns_
    assert "a_comp_cor_01" in conf.columns_
    assert "a_comp_cor_02" not in conf.columns_


def test_n_motion():

    conf = lc.Confounds(strategy=["motion"], motion="full", n_motion=0.2)
    conf.load(file_confounds)
    assert "motion_pca_1" in conf.columns_
    assert "motion_pca_2" not in conf.columns_

    conf = lc.Confounds(strategy=["motion"], motion="full", n_motion=0.95)
    conf.load(file_confounds)
    assert "motion_pca_6" in conf.columns_

    with pytest.raises(ValueError):
        conf = lc.Confounds(strategy=["motion"], motion="full", n_motion=50)
        conf.load(file_confounds)


def test_not_found_exception():

    conf = lc.Confounds(
        strategy=["high_pass", "motion", "global"], global_signal="full", motion="full"
    )

    missing_params = ["trans_y", "trans_x_derivative1", "rot_z_power2"]
    missing_keywords = ["cosine"]

    file_missing_confounds = os.path.join(
        path_data, "missing_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
    )

    with pytest.raises(ValueError) as exc_info:
        conf.load(file_missing_confounds)
    assert f"{missing_params}" in exc_info.value.args[0]
    assert f"{missing_keywords}" in exc_info.value.args[0]

    # loading anat compcor should also raise an error, because the json file is
    # missing for that example dataset
    with pytest.raises(ValueError):
        conf = lc.Confounds(strategy=["compcor"], compcor="anat")
        conf.load(file_missing_confounds)

    # catch invalid compcor option
    with pytest.raises(KeyError):
        conf = lc.Confounds(strategy=["compcor"], compcor="blah")
        conf.load(file_confounds)

    # catch invalid compcor option
    with pytest.raises(ValueError):
        conf = lc.Confounds(
            strategy=["compcor"], compcor="full", acompcor_combined=None
        )
        conf.load(file_confounds)

    # Aggressive ICA-AROMA strategy requires
    # default nifti and noise ICs in confound file
    # correct nifti but missing noise regressor
    with pytest.raises(ValueError) as exc_info:
        conf = lc.Confounds(strategy=["ica_aroma"], ica_aroma="basic")
        conf.load(file_missing_confounds)
    assert "aroma" in exc_info.value.args[0]

    # Aggressive ICA-AROMA strategy requires
    # default nifti
    aroma_nii = os.path.join(
        path_data, "test_space-MNI152NLin2009cAsym_desc-smoothAROMAnonaggr_bold.nii.gz"
    )
    with pytest.raises(ValueError) as exc_info:
        conf.load(aroma_nii)
    assert "Invalid file type" in exc_info.value.args[0]

    # non aggressive ICA-AROMA strategy requires
    # desc-smoothAROMAnonaggr nifti file
    with pytest.raises(ValueError) as exc_info:
        conf = lc.Confounds(strategy=["ica_aroma"], ica_aroma="full")
        conf.load(file_missing_confounds)
    assert "desc-smoothAROMAnonaggr_bold" in exc_info.value.args[0]


def test_load_non_nifti():
    """Test non-nifti and invalid file type as input."""
    conf = lc.Confounds()

    # tsv file - unsupported input
    tsv = os.path.join(path_data, "test_desc-confounds_regressors.tsv")
    with pytest.raises(ValueError):
        conf.load(tsv)

    # cifti file should be supported
    cifti = os.path.join(path_data, "test_space-fsLR_den-91k_bold.dtseries.nii")
    conf.load(cifti)
    assert conf.confounds_.size != 0

    # gifti support
    gifti = [
        os.path.join(path_data, f"test_space-fsaverage5_hemi-{hemi}_bold.func.gii")
        for hemi in ["L", "R"]
    ]
    conf.load(gifti)
    assert conf.confounds_.size != 0


def test_invalid_filetype():
    """Invalid file types/associated files for load method."""
    # invalid fmriprep version: contain confound files before and after v20.2.0
    conf = lc.Confounds()

    invalid_ver = os.path.join(
        path_data, "invalid_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
    )
    with pytest.raises(ValueError):
        conf.load(invalid_ver)

    # nifti with no associated confound file
    no_confound = os.path.join(
        path_data, "noconfound_space-MNI152NLin2009cAsym_desc-preproc_bold.nii.gz"
    )
    with pytest.raises(ValueError):
        conf.load(no_confound)


def test_ica_aroma():
    """Test ICA AROMA related file input."""
    aroma_nii = os.path.join(
        path_data, "test_space-MNI152NLin2009cAsym_desc-smoothAROMAnonaggr_bold.nii.gz"
    )
    # Agressive strategy
    conf = lc.Confounds(strategy=["ica_aroma"], ica_aroma="basic")
    conf.load(file_confounds)
    for col_name in conf.columns_:
        assert re.match("aroma_motion_+", col_name)

    # Non-agressive strategy
    conf = lc.Confounds(strategy=["ica_aroma"], ica_aroma="full")
    conf.load(aroma_nii)
    assert conf.confounds_.size == 0

    # invalid combination of strategy and option
    with pytest.raises(ValueError) as exc_info:
        conf = lc.Confounds(strategy=["ica_aroma"], ica_aroma=None)
        conf.load(file_confounds)
    assert "ICA-AROMA strategy" in exc_info.value.args[0]



from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="load_confounds",
    version="0.12.0",
    description="load fMRIprep confounds in python",
    license="MIT",
    url="https://github.com/simexp/load_confounds",
    long_description=long_description,
    long_description_content_type="text/markdown",  # Optional (see note above)
    project_urls={  # Optional
        "Bug Reports": "https://github.com/simexp/load_confounds/issues",
        "Funding": "https://cneuromod.ca",
        "Source": "https://github.com/simexp/load_confounds",
    },
    maintainer="Pierre Bellec",
    maintainer_email="pierre.bellec@gmail.com",
    packages=find_packages(),
    package_data={"load_confounds.data": ["*.nii.gz", "*.tsv"]},
    install_requires=[
        "numpy>=1.17.4",
        "pandas>=0.25.3",
        "scikit-learn>=0.21.3",
        "scipy>=1.3.2",
        "nilearn>=0.7.1",
    ],  # external packages as dependencies
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires=">=3.5",
)



"""Helper functions for the manipulation of confounds.

Authors: load_confounds team
"""
import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import scale
import os
import json
import re


img_file_patterns = {
    "aroma": "_desc-smoothAROMAnonaggr_bold",
    "nii.gz": "_space-.*_desc-preproc_bold.nii.gz",
    "dtseries.nii": "_space-.*_bold.dtseries.nii",
    "func.gii": "_space-.*_hemi-[LR]_bold.func.gii",
}

img_file_error = {
    "aroma": "Input must be ~desc-smoothAROMAnonaggr_bold for full ICA-AROMA strategy.",
    "nii.gz": "Invalid file type for the selected method.",
    "dtseries.nii": "Invalid file type for the selected method.",
    "func.gii": "need fMRIprep output with extension func.gii",
}


def _check_params(confounds_raw, params):
    """Check that specified parameters can be found in the confounds."""
    not_found_params = []
    for par in params:
        if not par in confounds_raw.columns:
            not_found_params.append(par)
    if not_found_params:
        raise MissingConfound(params=not_found_params)
    return None


def _find_confounds(confounds_raw, keywords):
    """Find confounds that contain certain keywords."""
    list_confounds = []
    missing_keys = []
    for key in keywords:
        key_found = [col for col in confounds_raw.columns if key in col]
        if not key_found:
            missing_keys.append(key)
        else:
            list_confounds.extend(key_found)
    if missing_keys:
        raise MissingConfound(keywords=missing_keys)
    return list_confounds


def _flag_single_gifti(img_files):
    """Test if the paired input files are giftis."""
    flag_single_gifti = []  # gifti in pairs
    for img in img_files:
        ext = ".".join(img.split(".")[-2:])
        flag_single_gifti.append((ext == "func.gii"))
    return all(flag_single_gifti)


def _sanitize_confounds(img_files):
    """Make sure the inputs are in the correct format."""
    # we want to support loading a single set of confounds, instead of a list
    # so we hack it
    if isinstance(img_files, list) and len(img_files) == 2:
        flag_single = _flag_single_gifti(img_files)
    else:  # single file
        flag_single = isinstance(img_files, str)

    if flag_single:
        img_files = [img_files]
    return img_files, flag_single


def _add_suffix(params, model):
    """
    Add suffixes to a list of parameters.
    Suffixes includes derivatives, power2 and full
    """
    params_full = params.copy()
    suffix = {
        "basic": {},
        "derivatives": {"derivative1"},
        "power2": {"power2"},
        "full": {"derivative1", "power2", "derivative1_power2"},
    }
    for par in params:
        for suff in suffix[model]:
            params_full.append(f"{par}_{suff}")
    return params_full


def _pca_motion(confounds_motion, n_components):
    """Reduce the motion paramaters using PCA."""
    n_available = confounds_motion.shape[1]
    if n_components > n_available:
        raise ValueError(
            f"User requested n_motion={n_components} motion components, but found only {n_available}."
        )
    confounds_motion = confounds_motion.dropna()
    confounds_motion_std = scale(
        confounds_motion, axis=0, with_mean=True, with_std=True
    )
    pca = PCA(n_components=n_components)
    motion_pca = pd.DataFrame(pca.fit_transform(confounds_motion_std))
    motion_pca.columns = ["motion_pca_" + str(col + 1) for col in motion_pca.columns]
    return motion_pca


def _optimize_scrub(fd_outliers, n_scans):
    """
    Perform optimized scrub. After scrub volumes, further remove
    continuous segments containing fewer than 5 volumes.
    Power, Jonathan D., et al. "Methods to detect, characterize, and remove
    motion artifact in resting state fMRI." Neuroimage 84 (2014): 320-341.
    """
    # Start by checking if the beginning continuous segment is fewer than 5 volumes
    if fd_outliers[0] < 5:
        fd_outliers = np.asarray(list(range(fd_outliers[0])) + list(fd_outliers))
    # Do the same for the ending segment of scans
    if n_scans - (fd_outliers[-1] + 1) < 5:
        fd_outliers = np.asarray(
            list(fd_outliers) + list(range(fd_outliers[-1], n_scans))
        )
    # Now do everything in between
    fd_outlier_ind_diffs = np.diff(fd_outliers)
    short_segments_inds = np.where(
        np.logical_and(fd_outlier_ind_diffs > 1, fd_outlier_ind_diffs < 6)
    )[0]
    for ind in short_segments_inds:
        fd_outliers = np.asarray(
            list(fd_outliers) + list(range(fd_outliers[ind] + 1, fd_outliers[ind + 1]))
        )
    fd_outliers = np.sort(np.unique(fd_outliers))
    return fd_outliers


def _get_file_raw(nii_file):
    """Get the name of the raw confound file."""
    if isinstance(nii_file, list):  # catch gifti
        nii_file = nii_file[0]
    suffix = "_space-" + nii_file.split("space-")[1]
    # fmriprep has changed the file suffix between v20.1.1 and v20.2.0 with respect to BEP 012.
    # cf. https://neurostars.org/t/naming-change-confounds-regressors-to-confounds-timeseries/17637
    # Check file with new naming scheme exists or replace, for backward compatibility.
    confounds_raw_candidates = [
        nii_file.replace(
            suffix,
            "_desc-confounds_timeseries.tsv",
        ),
        nii_file.replace(
            suffix,
            "_desc-confounds_regressors.tsv",
        ),
    ]

    confounds_raw = [cr for cr in confounds_raw_candidates if os.path.exists(cr)]

    if not confounds_raw:
        raise ValueError("Could not find associated confound file.")
    elif len(confounds_raw) != 1:
        raise ValueError("Found more than one confound file.")
    else:
        return confounds_raw[0]


def _get_json(confounds_raw, flag_acompcor):
    """Load json data companion to the confounds tsv file."""
    # Load JSON file
    confounds_json = confounds_raw.replace("tsv", "json")
    try:
        with open(confounds_json, "rb") as f:
            confounds_json = json.load(f)
    except OSError:
        if flag_acompcor:
            raise ValueError(
                f"Could not find a json file {confounds_json}. This is necessary for anat compcor"
            )
    return confounds_json


def _ext_validator(image_file, ext):
    """Check image is valid based on extention."""
    try:
        valid_img = all(
            bool(re.search(img_file_patterns[ext], img)) for img in image_file
        )
        error_message = img_file_error[ext]
    except KeyError:
        valid_img = False
        error_message = "Unsupported input."
    return valid_img, error_message


def _check_images(image_file, flag_full_aroma):
    """Validate input file and ICA AROMA related file."""
    if len(image_file) == 2:  # must be gifti
        valid_img, error_message = _ext_validator(image_file, "func.gii")
    elif flag_full_aroma:
        valid_img, error_message = _ext_validator([image_file], "aroma")
    else:
        ext = ".".join(image_file.split(".")[-2:])
        valid_img, error_message = _ext_validator([image_file], ext)
    if not valid_img:
        raise ValueError(error_message)


def _confounds_to_df(image_file, flag_acompcor, flag_full_aroma):
    """Load raw confounds as a pandas DataFrame."""
    _check_images(image_file, flag_full_aroma)
    confounds_raw = _get_file_raw(image_file)
    confounds_json = _get_json(confounds_raw, flag_acompcor)
    confounds_raw = pd.read_csv(confounds_raw, delimiter="\t", encoding="utf-8")
    return confounds_raw, confounds_json


def _confounds_to_ndarray(confounds, demean):
    """Convert confounds from a pandas dataframe to a numpy array."""
    # Convert from DataFrame to numpy ndarray
    labels = confounds.columns
    confounds = confounds.values

    # Derivatives have NaN on the first row
    # Replace them by estimates at second time point,
    # otherwise nilearn will crash.
    if confounds.size != 0:  # ica_aroma = "full" generate empty output
        mask_nan = np.isnan(confounds[0, :])
        confounds[0, mask_nan] = confounds[1, mask_nan]

        # Optionally demean confounds
        if demean:
            confounds = scale(confounds, axis=0, with_std=False)

    return confounds, labels


class MissingConfound(Exception):
    """
    Exception raised when failing to find params in the confounds.

    Parameters
    ----------
        params : list of missing params
        keywords: list of missing keywords
    """

    def __init__(self, params=None, keywords=None):
        """Default values are empty lists."""
        self.params = params if params else []
        self.keywords = keywords if keywords else []
